/*	$Id: hueblobMainCodels.cc 2007/12/01 21:16:01 mallet $ */

/*
 * Copyright (c) 2006 Anthony Mallet                  --  Wed Oct  4 2006
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice, this list of conditions and the following disclaimer.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice,  this list of  conditions and the following disclaimer in
 *      the  documentation  and/or  other   materials provided  with  the
 *      distribution.
 *
 * THIS  SOFTWARE IS PROVIDED BY  THE  COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND  ANY  EXPRESS OR IMPLIED  WARRANTIES,  INCLUDING,  BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES  OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR  PURPOSE ARE DISCLAIMED. IN  NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR      CONTRIBUTORS  BE LIABLE FOR   ANY    DIRECT, INDIRECT,
 * INCIDENTAL,  SPECIAL,  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN  CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE   OF THIS SOFTWARE, EVEN   IF ADVISED OF   THE POSSIBILITY OF SUCH
 * DAMAGE.
 */

#include <portLib.h>
#include <stdlib.h>
#include <err.h>
#include "server/hueblobHeader.h"

#include <cv.h>
#include <highgui.h>

#include "t3d/t3d.h"
#include "t3d/v3d.h"

#include "shared.h"


/* --- local data ---------------------------------------------------- */

#define VIAM_POSTER_NAME	"viam%s"

static char cameraPosterName[HUEBLOB_MAX_STRLEN];
static POSTER_ID cameraPosterId;	/* current camera poster */
static ViamImageBank *cameraImage;	/* current poster addr */
static int cameraImageNum;

static POSTER_ID spixPosterId;		/* current 3d image poster */
static Spix3DImage *spixImage;		/* current poster addr */
static POM_SENSOR_POS imagePos;		/* current image position */

static struct HueBlobObj {
  const char *name;
  int nViews;
  CvHistogram **modelHistogram;
} *objects;
static int nObjects;

static IplImage *trackImage, *hstrackImage[2], *trackBackProj, *thrBackProj;
static Spix3DImage *track3DImage;

static STATUS	getVideoImage(int *report);
static STATUS	get3DImage(int *report);
static STATUS	get3DBox(IplImage *bp, HueBlobImageBox *video, HueBlob3DBox *space);

static struct HueBlobObj *	getObject(const char *name);
static struct HueBlobObj *	addObject(const char *name);


/*------------------------------------------------------------------------
 *
 * hueblobMainInit  --  Initialization codel (fIDS, ...)
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
hueblobMainInit(int *report)
{
  objects = NULL;
  nObjects = 0;

  return OK;
}

/*------------------------------------------------------------------------
 *
 * hueblobMainEnd  --  Termination codel
 *
 * Description:
 *
 * Returns:    OK or ERROR
 */

STATUS
hueblobMainEnd(void)
{
  free(objects);
  return OK;
}


/* --- hueblobMainPerm ----------------------------------------------- */

/** Codel of permanent activity
 *
 * \return	OK or ERROR
 */

STATUS
hueblobMainPerm(int *report)
{
#ifdef HUEBLOB_DISPLAY
  if (SDI_F->display == HUEBLOB_ENABLE) {
    cvWaitKey(5);
  }
#endif
  return OK;
}


/* --- SetObject ----------------------------------------------------- */

/** Load object template image
 *
 * \param[in]	file	image file name
 * \param[out]	report	OK,
 *			S_hueblob_NO_SUCH_FILE,
 *			S_hueblob_IMAGE_UNREADABLE
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
hueblobSetObjectExec(HueBlobObject *oname, int *report)
{
  struct HueBlobObj *object;
  CvHistogram **objHist;
  IplImage *model;
  IplImage *hsv, *hs_planes[2];
  float max;

  int hist_size[] = {25, 25};
  float hue_range[] = { 0, 250 }; /* 0 (~0red) to 180 (~360red again) */
  float sat_range[] = { 0, 250 }; /* 0 (black-gray-white) to 255 (pure spectrum color) */
  float *hist_ranges[] = { hue_range, sat_range };

  /* read image */
  fprintf(stderr, "loading %s\n", oname->file);
  model = cvLoadImage(oname->file, 1);
  if (!model) {
    *report = S_hueblob_IMAGE_UNREADABLE;
    return ETHER;
  }

  /* create object structure */
  object = getObject(oname->object);
  if (!object) {
    object = addObject(oname->object);
    if (!object) {
      *report = S_hueblob_NO_SUCH_FILE;
      return ETHER;
    }
  }
  object->nViews++;
  object->modelHistogram =
    (CvHistogram **)realloc(object->modelHistogram,
			    object->nViews * sizeof(*object->modelHistogram));
  objHist = &object->modelHistogram[object->nViews-1];

  /* compute histogram */
  hsv = cvCreateImage(cvGetSize(model), 8, 3);
  hs_planes[0] = cvCreateImage(cvGetSize(model), 8, 1);
  hs_planes[1] = cvCreateImage(cvGetSize(model), 8, 1);

  cvCvtColor(model, hsv, CV_BGR2HSV);
  cvCvtPixToPlane(hsv, hs_planes[0], NULL, NULL, NULL);
  cvCvtPixToPlane(hsv, NULL, hs_planes[1], NULL, NULL);

  *objHist = cvCreateHist(2, hist_size, CV_HIST_ARRAY, hist_ranges, 1);
  cvCalcHist(hs_planes, *objHist, 0);
  cvGetMinMaxHistValue(*objHist, 0, &max, 0, 0 );
  cvConvertScale((*objHist)->bins, (*objHist)->bins, max?255./max:0., 0);

  cvReleaseImage(&hs_planes[1]);
  cvReleaseImage(&hs_planes[0]);
  cvReleaseImage(&hsv);
  cvReleaseImage(&model);
  return ETHER;
}


/* --- SetImage ------------------------------------------------------ */

/** Set image source and size
 *
 * \param[in]	image	image poster reference
 * \param[out]	report	OK,
 *			S_hueblob_LOW_MEMORY,
 *			S_hueblob_BAD_IMAGE,
 *			S_hueblob_NEED_COLOR_IMAGE,
 *			S_hueblob_NOT_FOUND
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
hueblobSetImageExec(HueBlobInput *image, int *report)
{
  int width, height;
  int i;

  sprintf(cameraPosterName, VIAM_POSTER_NAME, image->bank);
  fprintf(stderr, "looking for poster %s\n", cameraPosterName);

  if (posterFind(cameraPosterName, &cameraPosterId) != OK) {
    *report = S_hueblob_NOT_FOUND;
    return ETHER;
  }

  if (posterTake(cameraPosterId, POSTER_READ) != OK) {
    *report = errnoGet();
    return ETHER;
  }

  cameraImage = (ViamImageBank *)posterAddr(cameraPosterId);
  if (cameraImage == NULL) {
    posterGive(cameraPosterId);
    *report = S_hueblob_NOT_FOUND;
    return ETHER;
  }

  cameraImageNum = -1;
  for(i=0; i<cameraImage->nImages; i++)
    if (!strcmp(cameraImage->image[i].name.id, image->camera)) {
      cameraImageNum = i;
      break;
    }
  if (cameraImageNum < 0) {
    posterGive(cameraPosterId);
    *report = S_hueblob_BAD_IMAGE;
    return ETHER;
  }

  if (cameraImage->image[cameraImageNum].nChannels != 3) {
    posterGive(cameraPosterId);
    *report = S_hueblob_NEED_COLOR_IMAGE;
    return ETHER;
  }

  width = cameraImage->image[cameraImageNum].width;
  height = cameraImage->image[cameraImageNum].height;
  posterGive(cameraPosterId);

  fprintf(stderr, "looking for poster " STEREOPIXEL_3DIMAGE_POSTER_NAME "\n");
  if (posterFind(STEREOPIXEL_3DIMAGE_POSTER_NAME, &spixPosterId) != OK) {
    spixImage = NULL;
    fprintf(stderr,
	    "poster " STEREOPIXEL_3DIMAGE_POSTER_NAME " not found, using 2D only.\n");
  } else
    spixImage = (Spix3DImage *)posterAddr(spixPosterId);

  if (trackImage) cvReleaseImage(&trackImage);
  if (hstrackImage[0]) cvReleaseImage(&hstrackImage[0]);
  if (hstrackImage[1]) cvReleaseImage(&hstrackImage[1]);
  if (trackBackProj) cvReleaseImage(&trackBackProj);
  if (thrBackProj) cvReleaseImage(&thrBackProj);

  trackImage = cvCreateImage(cvSize(width, height), 8, 3);
  hstrackImage[0] = cvCreateImage(cvSize(width, height), 8, 1);
  hstrackImage[1] = cvCreateImage(cvSize(width, height), 8, 1);
  trackBackProj = cvCreateImage(cvSize(width, height), 8, 1);
  thrBackProj = cvCreateImage(cvSize(width, height), 8, 1);
  if (!trackImage || !hstrackImage[0] || !hstrackImage[1] || !trackBackProj) {
    *report = S_hueblob_LOW_MEMORY;
    return ETHER;
  }

  if (spixImage) {
    if (track3DImage) free(track3DImage);
    track3DImage = (Spix3DImage *)
      malloc(STEREOPIXEL_3DIMAGE_POSTER_SIZE(width, height));
    if (!track3DImage) {
      *report = S_hueblob_LOW_MEMORY;
      return ETHER;
    }
  } else
    track3DImage = NULL;

  fprintf(stderr, "using %dx%d image number %d\n",
	  width, height, cameraImageNum);
  return ETHER;
}


/* --- GetBox -------------------------------------------------------- */

/** Find object template in an image
 *
 * \param[out]	report	OK,
 *			S_hueblob_NO_MODEL,
 *			S_hueblob_NOT_FOUND
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
hueblobGetBlobExec(HueBlobImageBox *video, HueBlobBox *box, int *report)
{
  struct HueBlobObj *objmodel;
  CvTermCriteria criteria = cvTermCriteria(CV_TERMCRIT_ITER|CV_TERMCRIT_EPS, 50, 1);
  CvConnectedComp components;
  CvBox2D object;
  unsigned char *s, *d;
  unsigned int p;
  int i, nmodel;

  objmodel = getObject(video->object);
  if (!objmodel || objmodel->nViews == 0) {
    *report = S_hueblob_NO_MODEL;
    return ETHER;
  }

  if (getVideoImage(report) != OK) return ETHER;
  if (spixImage)
    if (get3DImage(report) != OK) return ETHER;

  /* reset search zone if it is incorrect */
  if (video->x < 0 || video->y < 0) {
    video->x = 0;
    video->y = 0;
    video->width = thrBackProj->width-1;
    video->height = thrBackProj->height-1;
  }
  if (video->x+video->width > thrBackProj->width-1)
    video->width = thrBackProj->width-1 - video->x;
  if (video->y+video->height > thrBackProj->height-1)
    video->height = thrBackProj->height-1 - video->y;

  /* convert source to HSV */
  cvCvtColor(trackImage, trackImage, CV_BGR2HSV);
  cvCvtPixToPlane(trackImage, hstrackImage[0], NULL, NULL, NULL);
  cvCvtPixToPlane(trackImage, NULL, hstrackImage[1], NULL, NULL);

  /* iterate over all templates */
  if (objmodel->nViews <= 1) {

    cvCalcBackProject(hstrackImage, thrBackProj, objmodel->modelHistogram[0]);

  } else {

    memset(thrBackProj->imageData, 0, thrBackProj->imageSize);
    for(nmodel = 0; nmodel<objmodel->nViews; nmodel++) {
      cvCalcBackProject(hstrackImage, trackBackProj, objmodel->modelHistogram[nmodel]);

      s = (unsigned char *)trackBackProj->imageData;
      d = (unsigned char *)thrBackProj->imageData;
      for(i=0;i<thrBackProj->imageSize;i++) {
	p = *s + *d; if (p>255) p = 255;
	*d = p;

	s++; d++;
      }
    }
  }

  cvThreshold(thrBackProj, trackBackProj, 32, 0, CV_THRESH_TOZERO);
  cvSmooth(trackBackProj, thrBackProj, CV_MEDIAN, 3);

  i = cvCamShift(thrBackProj,
		 cvRect(video->x, video->y, video->width, video->height),
		 criteria, &components, &object);
  if (object.size.height == 0 || object.size.width == 0) {
    box->video.x = box->video.y = -1;
    fprintf(stderr, "lost target\n");
    *report = S_hueblob_NOT_FOUND;
    return ETHER;
  }

  box->video.x = components.rect.x;
  box->video.y = components.rect.y;
  box->video.width = components.rect.width;
  box->video.height = components.rect.height;

  if (spixImage)
    if (get3DBox(thrBackProj, &box->video, &box->space) != OK) {
      fprintf(stderr, "no 3d coords\n");
      *report = S_hueblob_NOT_FOUND;
    }

#ifdef HUEBLOB_DISPLAY
  if (SDI_F->display == HUEBLOB_ENABLE) {

    cvNamedWindow("Source image", 0);
    cvResizeWindow("Source image", 320, 240);
    cvRectangle(trackImage,
		cvPoint(box->video.x, box->video.y),
		cvPoint(box->video.x+box->video.width,
			box->video.y+box->video.height),
		cvScalar(255), 1);
    cvShowImage("Source image", trackImage);

    cvNamedWindow("Back projection image", 0);
    cvResizeWindow("Back projection image", 320, 240);
    cvRectangle(thrBackProj,
		cvPoint(box->video.x, box->video.y),
		cvPoint(box->video.x+box->video.width,
			box->video.y+box->video.height),
		cvScalar(255), 1);
#if 0
    cvEllipse(thrBackProj,
	      cvPoint((int)(object.center.x+0.5), (int)(object.center.y+0.5)),
	      cvSize((int)object.size.height/2, (int)object.size.width/2),
	      object.angle, 0, 360,
	      cvScalar(255), 1);
#endif
    cvShowImage("Back projection image", thrBackProj);
  }
#endif

  return ETHER;
}


/* --- GetBoxLoop ---------------------------------------------------- */

/** Find object template repeatedly
 *
 * \param[out] report	OK,
 *			S_hueblob_NO_MODEL
 *			S_hueblob_NOT_FOUND
 *
 * \return EXEC END ETHER FAIL ZOMBIE
 */

ACTIVITY_EVENT
hueblobGetBoxLoopExec(HueBlobImageBox *video, HueBlobBox *box, int *report)
{
  *report = OK;
  video->nIter--;
  hueblobGetBlobExec(video, box, report);

  if (*report == OK) {
    video->x = box->video.x;
    video->y = box->video.y;
    video->width = box->video.width;
    video->height = box->video.height;
  } else {
    video->x = -1;
    video->y = -1;
    video->width = 0;
    video->height = 0;
    if (video->nIter > 0) *report = OK;
  }

  if (video->nIter <= 0) return ETHER;
  return EXEC;
}


/* --- TrackBlob ----------------------------------------------------- */

/** Perform detection and 3D localization of an object
 *
 * Reports:      OK
 *              S_hueblob_NOT_FOUND
 */

static POSTER_ID blobImagePosterId;	/**< chessboard viam poster */
static ViamImageBank *blobImage;	/**< chessboard poster addr */
static POSTER_ID blobCoordPosterId;	/**< chessboard output poster */
static T3D *blobCoord;			/**< chessboard location */

static IplImage *blobtrackImage[3];	/**< tracked images */

/* hueblobTrackChessBoardStart  -  codel START of TrackChessBoard
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
hueblobTrackBlobStart(HueBlobTrackBlob *blob, int *report)
{
  char name[HUEBLOB_MAX_STRLEN];
  STATUS s;
  int i;

  /* --- get image bank --- */

  sprintf(name, VIAM_POSTER_NAME, blob->bank);
  fprintf(stderr, "looking for poster %s\n", name);

  if (posterFind(name, &blobImagePosterId) != OK) {
    *report = S_hueblob_NOT_FOUND;
    return ETHER;
  }

  if (posterTake(blobImagePosterId, POSTER_READ) != OK) {
    *report = errnoGet();
    return ETHER;
  }

  blobImage = (ViamImageBank *)posterAddr(blobImagePosterId);
  if (blobImage == NULL) {
    posterGive(blobImagePosterId);
    *report = S_hueblob_NOT_FOUND;
    return ETHER;
  }
  if (blobImage->nImages != 2) {
    posterGive(blobImagePosterId);
    *report = S_hueblob_NOT_STEREO_BANK;
    return ETHER;
  }

  blobtrackImage[0] = cvCreateImage(
    cvSize(blobImage->image[0].width, blobImage->image[0].height),
    8, 3);
  for(i=0; i<2; i++) {
    blobtrackImage[i+1] = cvCreateImage(
      cvSize(blobImage->image[i].width, blobImage->image[i].height),
      8, 1);
  }

  posterGive(blobImagePosterId);

  /* --- create output poster --- */

  s = posterCreate(blob->name, sizeof(*blobCoord), &blobCoordPosterId);
  if (s != OK) { *report = errnoGet(); return ETHER; }
  blobCoord = (T3D *)posterAddr(blobCoordPosterId);

  posterTake(blobCoordPosterId, POSTER_WRITE);
  memset(blobCoord, 0, sizeof(*blobCoord));
  posterGive(blobCoordPosterId);

  return EXEC;
}

/* hueblobTrackChessBoardMain  -  codel EXEC of TrackChessBoard
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
hueblobTrackBlobMain(HueBlobTrackBlob *blob, int *report)
{
  struct HueBlobObj *objmodel;
  CvTermCriteria criteria = cvTermCriteria(CV_TERMCRIT_ITER|CV_TERMCRIT_EPS, 50, 1);
  CvConnectedComp components;
  CvBox2D object;
  unsigned char *s, *d;
  unsigned int p;
  int nmodel;
  double intrinsic[9], baseline;
  IplImage viam;
  T3D pose;
  int i;
  DATA_IM3D im3d;
  T3D c2w;

  /* --- retrieve video image --- */

  if (posterTake(blobImagePosterId, POSTER_READ) != OK) {
    *report = errnoGet();
    return ETHER;
  }
  baseline = blobImage->calibration.baseline;
  memcpy(intrinsic, blobImage->image[0].calibration.intrirect,
	 sizeof(intrinsic));
  c2w = t3dBryanIdentity;
  c2w.bryan.yaw = blobImage->image[0].pos.sensorToMain.euler.yaw;
  c2w.bryan.pitch = blobImage->image[0].pos.sensorToMain.euler.pitch;
  c2w.bryan.roll = blobImage->image[0].pos.sensorToMain.euler.roll;
  c2w.bryan.x = blobImage->image[0].pos.sensorToMain.euler.x;
  c2w.bryan.y = blobImage->image[0].pos.sensorToMain.euler.y;
  c2w.bryan.z = blobImage->image[0].pos.sensorToMain.euler.z;
  c2w.flags |= T3D_ALLOW_CONVERSION;
  for(i=0; i<2; i++) {
    cvInitImageHeader(&viam,
		      cvSize(blobImage->image[i].width,
			     blobImage->image[i].height),
		      blobImage->image[i].depth,
		      blobImage->image[i].nChannels,
		      IPL_ORIGIN_TL, 4);
    cvSetData(&viam,
	      &blobImage->image[i].data[blobImage->image[i].dataOffset],
	      blobImage->image[i].widthStep);
    cvCvtColor(&viam, blobtrackImage[i+1], CV_BGR2GRAY);
    if (i == 0)
      cvCvtColor(&viam, blobtrackImage[0], CV_BGR2HSV);
  }
  posterGive(blobImagePosterId);

  /* --- detect blob --- */

  objmodel = getObject(blob->object);
  if (!objmodel || objmodel->nViews == 0) {
    *report = S_hueblob_NO_MODEL;
    return ETHER;
  }

  /* reset search zone if it is incorrect */
  if (blob->x < 0 || blob->y < 0) {
    blob->x = 0;
    blob->y = 0;
    blob->width = thrBackProj->width-1;
    blob->height = thrBackProj->height-1;
  }
  if (blob->x+blob->width > thrBackProj->width-1)
    blob->width = thrBackProj->width-1 - blob->x;
  if (blob->y+blob->height > thrBackProj->height-1)
    blob->height = thrBackProj->height-1 - blob->y;

  /* convert source to HSV */
  cvCvtPixToPlane(blobtrackImage[0], hstrackImage[0], NULL, NULL, NULL);
  cvCvtPixToPlane(blobtrackImage[0], NULL, hstrackImage[1], NULL, NULL);

  /* iterate over all templates */
  if (objmodel->nViews <= 1) {

    cvCalcBackProject(hstrackImage, thrBackProj, objmodel->modelHistogram[0]);

  } else {

    memset(thrBackProj->imageData, 0, thrBackProj->imageSize);
    for(nmodel = 0; nmodel<objmodel->nViews; nmodel++) {
      cvCalcBackProject(hstrackImage, trackBackProj, objmodel->modelHistogram[nmodel]);

      s = (unsigned char *)trackBackProj->imageData;
      d = (unsigned char *)thrBackProj->imageData;
      for(i=0;i<thrBackProj->imageSize;i++) {
	p = *s + *d; if (p>255) p = 255;
	*d = p;

	s++; d++;
      }
    }
  }

  cvThreshold(thrBackProj, trackBackProj, 32, 0, CV_THRESH_TOZERO);
  cvSmooth(trackBackProj, thrBackProj, CV_MEDIAN, 3);

  i = cvCamShift(thrBackProj,
		 cvRect(blob->x, blob->y, blob->width, blob->height),
		 criteria, &components, &object);
  if (object.size.height == 0 || object.size.width == 0) {
    blob->x = blob->y = -1;
    warnx("cannot find blob");
#ifdef HUEBLOB_DISPLAY
    if (blob->feedback == HUEBLOB_ENABLE) {
	cvNamedWindow("Blob image", 0);
	cvResizeWindow("Blob image", 320, 240);
	cvShowImage("Blob image", blobtrackImage[0]);
    }
#endif
    return EXEC;
  }

  blob->x = components.rect.x;
  blob->y = components.rect.y;
  blob->width = components.rect.width;
  blob->height = components.rect.height;

#ifdef HUEBLOB_DISPLAY
  if (blob->feedback == HUEBLOB_ENABLE) {
    cvRectangle(blobtrackImage[0],
		cvPoint(blob->x, blob->y),
		cvPoint(blob->x+blob->width, blob->y+blob->height),
		cvScalar(255), 1);
    cvNamedWindow("Blob image", 0);
    cvResizeWindow("Blob image", 320, 240);
    cvShowImage("Blob image", blobtrackImage[0]);
  }
#endif /* HUEBLOB_DISPLAY */

  /* --- compute 3D position --- */

  {
    int off, hei;

    off = blob->y - 4; if (off < 0) off = 0;
    hei = blob->height + 4; if (off+hei > blobtrackImage[1]->height)
			      hei = blobtrackImage[1]->height - off;

  pose = t3dMatrixIdentity;

  if (init_data_im3d(&im3d,
		     hei, blobtrackImage[1]->width, 0,0,0) != ERR_IM3D_OK) {
    return EXEC;
  }

  {
    unsigned char *images[2];

    intrinsic[5] -= off;

    images[0] = (unsigned char*)blobtrackImage[1]->imageData + off*blobtrackImage[1]->widthStep;
    images[1] = (unsigned char*)blobtrackImage[2]->imageData + off*blobtrackImage[2]->widthStep;

    if (hb_correlation_compute(images, blobtrackImage[1]->width, hei,
			       baseline, intrinsic, &im3d, report) != OK) {
      posterTake(blobCoordPosterId, POSTER_WRITE);
      memset(blobCoord, 0, sizeof(*blobCoord));
      posterGive(blobCoordPosterId);
      empty_data_im3d(&im3d);
      return EXEC;
    }
  }

  {
    DATA_PT3D *b;
    unsigned char *p;
    double x, y, z;
    int i, j, n, t;

    /* compute average */
    x = y = z = 0;
    n = t = 0;
    b = im3d.points_tab + blob->x;
    p = (unsigned char *)thrBackProj->imageData +
      off * thrBackProj->widthStep + blob->x;
    for(i = 0; i < blob->height; i++) {
      for(j = 0; j < blob->width; j++) {
	if (b->state == GOOD_PT3D) {
	  x += b->coord_1 * (*p);
	  y += b->coord_2 * (*p);
	  z += b->coord_3 * (*p);
	  n += *p;
	  t++;
	}
	b++;
	p++;
      }
      b += im3d.header.nbcol - blob->width;
      p += thrBackProj->widthStep - blob->width;
    }

    fprintf(stderr, "using %d 3d points\n", t);
    if (n <= 0) {
      posterTake(blobCoordPosterId, POSTER_WRITE);
      memset(blobCoord, 0, sizeof(*blobCoord));
      posterGive(blobCoordPosterId);
      empty_data_im3d(&im3d);
      return EXEC;
    }

    /* XXX temporarily convert camera 3d frame to pom's one */
    pose.matrix.matrix[3] = z / n + 0.02;
    pose.matrix.matrix[7] = -x / n;
    pose.matrix.matrix[11] = -y / n;
  }}

  empty_data_im3d(&im3d);

  t3dComp(&pose, &c2w);
  posterTake(blobCoordPosterId, POSTER_WRITE);
  *blobCoord = pose;
  posterGive(blobCoordPosterId);

  if (blob->feedback == HUEBLOB_ENABLE)
    t3dFPrint(stdout, &pose);

  return EXEC;
}

/* hueblobTrackChessBoardEnd  -  codel END of TrackChessBoard
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
hueblobTrackBlobEnd(HueBlobTrackBlob *blob, int *report)
{
  int i;

  cvReleaseImage(&blobtrackImage[0]);
  for(i=0; i<2; i++) {
    cvReleaseImage(&blobtrackImage[i+1]);
  }
  posterDelete(blobCoordPosterId);
  return ETHER;
}


/* --- TrackChessBoard ----------------------------------------------- */

/** Perform detection and 3D localization of a chessboard pattern
 *
 * Reports:      OK
 *              S_hueblob_NOT_FOUND
 */

static POSTER_ID chessImagePosterId;	/**< chessboard viam poster */
static ViamImageBank *chessImage;	/**< chessboard poster addr */
static POSTER_ID chessCoordPosterId;	/**< chessboard output poster */
static T3D *chessCoord;			/**< chessboard location */

static IplImage *chesstrackImage[2];	/**< tracked images */
static CvMat *ipoints[2];		/**< 2d corners */

/* hueblobTrackChessBoardStart  -  codel START of TrackChessBoard
   Returns:  START EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
hueblobTrackChessBoardStart(HueBlobTrackChessboard *chessboard, int *report)
{
  char name[HUEBLOB_MAX_STRLEN];
  STATUS s;
  int i;

  /* --- get image bank --- */

  sprintf(name, VIAM_POSTER_NAME, chessboard->bank);
  fprintf(stderr, "looking for poster %s\n", name);

  if (posterFind(name, &chessImagePosterId) != OK) {
    *report = S_hueblob_NOT_FOUND;
    return ETHER;
  }

  if (posterTake(chessImagePosterId, POSTER_READ) != OK) {
    *report = errnoGet();
    return ETHER;
  }

  chessImage = (ViamImageBank *)posterAddr(chessImagePosterId);
  if (chessImage == NULL) {
    posterGive(chessImagePosterId);
    *report = S_hueblob_NOT_FOUND;
    return ETHER;
  }
  if (chessImage->nImages != 2) {
    posterGive(chessImagePosterId);
    *report = S_hueblob_NOT_STEREO_BANK;
    return ETHER;
  }

  for(i=0; i<2; i++) {
    chesstrackImage[i] = cvCreateImage(
      cvSize(chessImage->image[i].width, chessImage->image[i].height),
      8, 1);
  }

  posterGive(chessImagePosterId);

  /* --- create output poster --- */

  s = posterCreate(chessboard->name,
		   sizeof(*chessCoord), &chessCoordPosterId);
  if (s != OK) { *report = errnoGet(); return ETHER; }
  chessCoord = (T3D *)posterAddr(chessCoordPosterId);

  posterTake(chessCoordPosterId, POSTER_WRITE);
  memset(chessCoord, 0, sizeof(*chessCoord));
  posterGive(chessCoordPosterId);

  /* --- create points structures --- */

  chessboard->rows--;
  chessboard->cols--;

  for(i=0; i<2; i++)
    ipoints[i] = cvCreateMat(chessboard->rows*chessboard->cols, 2, CV_32FC1);

  return EXEC;
}

/* hueblobTrackChessBoardMain  -  codel EXEC of TrackChessBoard
   Returns:  EXEC END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
hueblobTrackChessBoardMain(HueBlobTrackChessboard *chessboard, int *report)
{
  double intrinsic[9], baseline;
  double points[chessboard->rows*chessboard->cols][3];
  IplImage viam;
  T3D pose;
  T3D c2w;
  double norm;
  int i, s, count;

  /* --- retrieve video image --- */

  if (posterTake(chessImagePosterId, POSTER_READ) != OK) {
    *report = errnoGet();
    return ETHER;
  }
  baseline = chessImage->calibration.baseline;
  baseline *= chessImage->image[0].calibration.intrirect[0];
  memcpy(intrinsic, chessImage->image[0].calibration.intrirect,
	 sizeof(intrinsic));
  c2w = t3dBryanIdentity;
  c2w.bryan.yaw = chessImage->image[0].pos.sensorToMain.euler.yaw;
  c2w.bryan.pitch = chessImage->image[0].pos.sensorToMain.euler.pitch;
  c2w.bryan.roll = chessImage->image[0].pos.sensorToMain.euler.roll;
  c2w.bryan.x = chessImage->image[0].pos.sensorToMain.euler.x;
  c2w.bryan.y = chessImage->image[0].pos.sensorToMain.euler.y;
  c2w.bryan.z = chessImage->image[0].pos.sensorToMain.euler.z;
  c2w.flags |= T3D_ALLOW_CONVERSION;

  for(i=0; i<2; i++) {
    cvInitImageHeader(&viam,
		      cvSize(chessImage->image[i].width,
			     chessImage->image[i].height),
		      chessImage->image[i].depth,
		      chessImage->image[i].nChannels,
		      IPL_ORIGIN_TL, 4);
    cvSetData(&viam,
	      &chessImage->image[i].data[chessImage->image[i].dataOffset],
	      chessImage->image[i].widthStep);
    cvCvtColor(&viam, chesstrackImage[i], CV_BGR2GRAY);
  }
  posterGive(chessImagePosterId);

  /* --- detect chessboard --- */

  for(i=0; i<2; i++) {
    s = cvFindChessboardCorners(chesstrackImage[i],
				cvSize(chessboard->cols, chessboard->rows),
				(CvPoint2D32f *)ipoints[i]->data.fl,
				&count,
				CV_CALIB_CB_ADAPTIVE_THRESH|
				CV_CALIB_CB_NORMALIZE_IMAGE);
    if (!s) {
      warnx("cannot get chessboard in image %d", i);
#ifdef HUEBLOB_DISPLAY
      if (chessboard->feedback == HUEBLOB_ENABLE) {
	cvDrawChessboardCorners(chesstrackImage[0],
				cvSize(chessboard->cols, chessboard->rows),
				(CvPoint2D32f *)ipoints[0]->data.fl, count, 0);
	cvNamedWindow("Chessboard image", 0);
	cvResizeWindow("Chessboard image", 320, 240);
	cvShowImage("Chessboard image", chesstrackImage[i]);
      }
#endif
      return EXEC;
    }

    cvFindCornerSubPix(chesstrackImage[i],
		       (CvPoint2D32f *)ipoints[i]->data.fl, count,
		       cvSize(5, 5), cvSize(-1, -1),
		       cvTermCriteria(CV_TERMCRIT_EPS, 100, 0.001));
  }

#ifdef HUEBLOB_DISPLAY
  if (chessboard->feedback == HUEBLOB_ENABLE) {
    cvDrawChessboardCorners(chesstrackImage[0],
			    cvSize(chessboard->cols, chessboard->rows),
			    (CvPoint2D32f *)ipoints[0]->data.fl, count, 1);
    cvNamedWindow("Chessboard image", 0);
    cvResizeWindow("Chessboard image", 320, 240);
    cvShowImage("Chessboard image", chesstrackImage[0]);
  }
#endif /* HUEBLOB_DISPLAY */

  /* --- compute 3D position --- */

  V3D x,y,z;
  pose = t3dMatrixIdentity;

  /* 3d coordinates of points */
  for(i=0; i<count; i++) {
    double disp;

    disp = ipoints[0]->data.fl[i*2] - ipoints[1]->data.fl[i*2];
    points[i][0] = baseline/disp;
    points[i][2] = -points[i][0]*
      (ipoints[0]->data.fl[i*2+1] - intrinsic[5])/intrinsic[4];
    points[i][1] = -points[i][0]*
      (ipoints[0]->data.fl[i*2]   - intrinsic[2])/intrinsic[0];
  }

  /* X vector */
  i = chessboard->cols - 1;
  x[0] = points[i][0] - points[0][0];
  x[1] = points[i][1] - points[0][1];
  x[2] = points[i][2] - points[0][2];
  v3dNormalize(&x);

  /* Y vector */
  i = (chessboard->rows - 1) * chessboard->cols;
  y[0] = points[i][0] - points[0][0];
  y[1] = points[i][1] - points[0][1];
  y[2] = points[i][2] - points[0][2];
  v3dNormalize(&y);

  /* Z vector */
  z[0] = x[1]*y[2] - x[2]*y[1];
  z[1] = x[2]*y[0] - x[0]*y[2];
  z[2] = x[0]*y[1] - x[1]*y[0];

  /* origin */
  for(i=0; i<count; i++) {
    pose.matrix.matrix[3] += points[i][0];
    pose.matrix.matrix[7] += points[i][1];
    pose.matrix.matrix[11] += points[i][2];
  }
  pose.matrix.matrix[3] /= count;
  pose.matrix.matrix[7] /= count;
  pose.matrix.matrix[11] /= count;

  warnx("%g %g %g", y[0], y[1], y[2]);

  /* 0.12 Y */
  pose.matrix.matrix[3] += 0.1 * y[0];
  pose.matrix.matrix[7] += 0.1 * y[1];
  pose.matrix.matrix[11] += 0.1 * y[2];
  /* 0.04 Z */
  pose.matrix.matrix[3] += 0.05 * z[0];
  pose.matrix.matrix[7] += 0.05 * z[1];
  pose.matrix.matrix[11] += 0.05 * z[2];

  t3dComp(&pose, &c2w);
  posterTake(chessCoordPosterId, POSTER_WRITE);
  *chessCoord = pose;
  posterGive(chessCoordPosterId);

  if (chessboard->feedback == HUEBLOB_ENABLE)
    t3dFPrint(stdout, &pose);

  return EXEC;
}

/* hueblobTrackChessBoardEnd  -  codel END of TrackChessBoard
   Returns:  END ETHER FAIL ZOMBIE */
ACTIVITY_EVENT
hueblobTrackChessBoardEnd(HueBlobTrackChessboard *chessboard, int *report)
{
  int i;

  for(i=0; i<2; i++) {
    cvReleaseMat(&ipoints[i]);
    cvReleaseImage(&chesstrackImage[i]);
  }
  posterDelete(chessCoordPosterId);
  return ETHER;
}


/* --- getVideoImage ------------------------------------------------- */

/** Update current video image from camera poster
 *
 * \param[out]	report	OK,
 *			S_hueblob_NOT_FOUND
 *
 * \return OK or ERROR
 */

static STATUS
getVideoImage(int *report)
{
  int i, j, delta;
  unsigned char *src, *dst;

  if (posterTake(cameraPosterId, POSTER_READ) != OK) {
    *report = errnoGet();
    return ERROR;
  }

  /* copy image */
  imagePos = cameraImage->image[cameraImageNum].pos;
  src = &cameraImage->image[cameraImageNum].data[
    cameraImage->image[cameraImageNum].dataOffset];
  dst = (unsigned char *)trackImage->imageData;
  delta = trackImage->widthStep - trackImage->width*3;
  for(i=0; i<cameraImage->image[cameraImageNum].height; i++) {
    for(j=0; j<cameraImage->image[cameraImageNum].width; j++) {
      /* copy BGR to BGR */
      *(dst++) = *(src++);
      *(dst++) = *(src++);
      *(dst++) = *(src++);
    }
    dst += delta;
  }

  posterGive(cameraPosterId);
  return OK;
}


/* --- get3DImage ---------------------------------------------------- */

/** Update current 3D image from stereopixel poster
 *
 * \param[out]	report	OK,
 *			S_hueblob_NOT_FOUND
 *
 * \return OK or ERROR
 */

static STATUS
get3DImage(int *report)
{
  int i, j;
  Spix3DPixel *src, *dst;

  if (posterTake(spixPosterId, POSTER_READ) != OK) {
    *report = errnoGet();
    return ERROR;
  }

  /* get position */
  imagePos = spixImage->imagePos;

  /* copy image */
  track3DImage->width = spixImage->width;
  track3DImage->height = spixImage->height;

  dst = track3DImage->pixel;
  src = spixImage->pixel;
  for(i=0; i<spixImage->width; i++) {
    for(j=0; j<spixImage->height; j++) {
      *(dst++) = *(src++);
    }
  }

  posterGive(spixPosterId);
  return OK;
}


/* --- get3DBox ------------------------------------------------------ */

/** Get 3d box volume
 *
 * \param[in] video	Box coordinates in video image
 * \param[out] space	Box coordinates in 3d space
 *
 * \return OK or ERROR
 */

static STATUS
get3DBox(IplImage *bp, HueBlobImageBox *video, HueBlob3DBox *space)
{
  T3D t3d;
  Spix3DPixel *b;
  unsigned char *p;
  double x, y, z;
  int i, j, n, t;

  /* compute average */
  x = y = z = 0;
  n = t = 0;
  b = track3DImage->pixel + video->y * track3DImage->width + video->x;
  p = (unsigned char *)bp->imageData + video->y * bp->widthStep + video->x;
  for(i = 0; i < video->height; i++) {
    for(j = 0; j < video->width; j++) {
      if (b->good) {
	x += b->x * (*p);
	y += b->y * (*p);
	z += b->z * (*p);
	n += *p;
	t++;
      }
      b++;
      p++;
    }
    b += track3DImage->width - video->width;
    p += bp->widthStep - video->width;
  }

  fprintf(stderr, "using %d 3d points, average weight %g\n", t, (double)n/t);
  fprintf(stderr, "average pos x %g y %g z %g\n", x/n, y/n, z/n);
  if (n <= 0) return ERROR;

  /* XXX temporarily convert camera 3d frame to pom's one */
  space->x = z / n;
  space->y = -x / n;
  space->z = -y / n;

  space->x += 0.015;

  /* change to 'main' frame. Who broke pom interaface?? There used
   * to be T3D in pom's posters, not that weird POM_SENSOR_POS. */
  t3d = t3dBryanIdentity;
  t3d.euler.euler[0] = imagePos.sensorToMain.euler.yaw;
  t3d.euler.euler[1] = imagePos.sensorToMain.euler.pitch;
  t3d.euler.euler[2] = imagePos.sensorToMain.euler.roll;
  t3d.euler.euler[3] = imagePos.sensorToMain.euler.x;
  t3d.euler.euler[4] = imagePos.sensorToMain.euler.y;
  t3d.euler.euler[5] = imagePos.sensorToMain.euler.z;
  t3dConvertTo(T3D_MATRIX, &t3d);

  x  = t3d.matrix.matrix[0] * space->x;
  x += t3d.matrix.matrix[1] * space->y;
  x += t3d.matrix.matrix[2] * space->z;
  x += t3d.matrix.matrix[3];

  y  = t3d.matrix.matrix[4] * space->x;
  y += t3d.matrix.matrix[5] * space->y;
  y += t3d.matrix.matrix[6] * space->z;
  y += t3d.matrix.matrix[7];

  z  = t3d.matrix.matrix[ 8] * space->x;
  z += t3d.matrix.matrix[ 9] * space->y;
  z += t3d.matrix.matrix[10] * space->z;
  z += t3d.matrix.matrix[11];

  space->x = x;
  space->y = y;
  space->z = z;

  return OK;
}


/* --- getObject ----------------------------------------------------- */

static struct HueBlobObj *
getObject(const char *name)
{
  struct HueBlobObj *o;
  int i;

  for(i=0, o=objects; i<nObjects; i++, o++)
    if (!strcmp(o->name, name)) return o;

  return NULL;
}


/* --- addObject ----------------------------------------------------- */

static struct HueBlobObj *
addObject(const char *name)
{
  objects = (HueBlobObj *)realloc(objects, (nObjects+1)*sizeof(*objects));

  objects[nObjects].name = strdup(name);
  objects[nObjects].nViews = 0;
  objects[nObjects].modelHistogram = NULL;

  return &objects[nObjects++];
}
